from __future__ import annotations

try:
    from importlib.resources import as_file, files
except ImportError:
    from importlib_resources import as_file, files  # type: ignore[import-not-found,no-redef]  # noqa: E501
from tempfile import NamedTemporaryFile
from unittest import mock

from brother_ql.backends.generic import BrotherQLBackendGeneric
from brother_ql.labels import FormFactor
from brother_ql.raster import BrotherQLRaster
from brother_ql_web import labels
from brother_ql_web.configuration import Font
from PIL import Image, ImageChops, ImageFont

from tests import TestCase


class LabelParametersTestCase(TestCase):
    def test_kind(self) -> None:
        parameters = labels.LabelParameters(
            configuration=self.example_configuration,
            label_size="38",
        )
        self.assertEqual(FormFactor.ENDLESS, parameters.kind)
        parameters.label_size = "62x29"
        self.assertEqual(FormFactor.DIE_CUT, parameters.kind)

    def test_scale_margin(self) -> None:
        parameters = labels.LabelParameters(
            configuration=self.example_configuration,
            font_size=37,
            margin_top=10,
            margin_bottom=25,
            margin_left=33,
            margin_right=57,
        )
        self.assertEqual(3, parameters.margin_top_scaled)  # 3.7
        self.assertEqual(9, parameters.margin_bottom_scaled)  # 9.25
        self.assertEqual(12, parameters.margin_left_scaled)  # 12.21
        self.assertEqual(21, parameters.margin_right_scaled)  # 21.09

    def test_fill_color(self) -> None:
        parameters = labels.LabelParameters(
            configuration=self.example_configuration,
            label_size="62",
        )
        self.assertEqual((0, 0, 0), parameters.fill_color)
        parameters.label_size = "62red"
        self.assertEqual((255, 0, 0), parameters.fill_color)

    def test_font_path(self) -> None:
        parameters = labels.LabelParameters(
            configuration=self.example_configuration,
            font_family=None,
            font_style=None,
        )
        parameters.configuration.label.default_font = Font(
            family="DejaVu Serif", style="Book"
        )

        # 1) Fallback to default.
        self.assertEqual(
            "/usr/share/fonts/truetype/dejavu/DejaVuSerif.ttf", parameters.font_path
        )

        # 2) Retrieve existing.
        parameters.font_family = "Roboto"
        parameters.font_style = "Medium"
        self.assertEqual(
            "/usr/share/fonts/truetype/roboto/unhinted/RobotoTTF/Roboto-Medium.ttf",
            parameters.font_path,
        )

        # 3) Retrieve missing.
        parameters.font_family = "Custom family"
        parameters.font_style = "Regular"
        with self.assertRaisesRegex(
            expected_exception=LookupError,
            expected_regex=r"^Couldn't find the font & style$",
        ):
            parameters.font_path

    def test_width_height(self) -> None:
        parameters = labels.LabelParameters(
            configuration=self.example_configuration,
        )

        # 1) Unknown label size.
        parameters.label_size = "1337"
        with self.assertRaisesRegex(
            expected_exception=LookupError, expected_regex=r"^Unknown label_size$"
        ):
            parameters.width_height

        # 2) Width > height. Handle standard and rotated.
        parameters.label_size = "62x29"
        self.assertEqual((696, 271), parameters.width_height)
        self.assertEqual(696, parameters.width)
        self.assertEqual(271, parameters.height)

        parameters.orientation = "rotated"
        self.assertEqual((271, 696), parameters.width_height)
        self.assertEqual(271, parameters.width)
        self.assertEqual(696, parameters.height)

        # 3) Height > width. Handle standard and rotated.
        parameters.label_size = "39x48"
        parameters.orientation = "standard"
        self.assertEqual((495, 425), parameters.width_height)
        self.assertEqual(495, parameters.width)
        self.assertEqual(425, parameters.height)

        parameters.orientation = "rotated"
        self.assertEqual((425, 495), parameters.width_height)
        self.assertEqual(425, parameters.width)
        self.assertEqual(495, parameters.height)

        # 4) Endless labels.
        parameters.label_size = "62"
        parameters.orientation = "standard"
        self.assertEqual((696, 0), parameters.width_height)
        self.assertEqual(696, parameters.width)
        self.assertEqual(0, parameters.height)

        parameters.orientation = "rotated"
        self.assertEqual((0, 696), parameters.width_height)
        self.assertEqual(0, parameters.width)
        self.assertEqual(696, parameters.height)


class DetermineImageDimensionsTestCase(TestCase):
    def test_determine_image_dimensions(self) -> None:
        parameters = labels.LabelParameters(
            configuration=self.example_configuration,
            font_family="Roboto",
            font_style="Medium",
        )
        image_font = ImageFont.truetype(parameters.font_path, parameters.font_size)
        text = "Test text"

        # 1) Fixed size labels.
        parameters.label_size = "62x29"
        parameters.orientation = "standard"
        result = labels._determine_image_dimensions(
            text=text, image_font=image_font, parameters=parameters
        )
        self.assertEqual((696, 271, 391, 72), result)

        parameters.orientation = "rotated"
        result = labels._determine_image_dimensions(
            text=text, image_font=image_font, parameters=parameters
        )
        self.assertEqual((271, 696, 391, 72), result)

        # 2) Endless labels.
        parameters.label_size = "62"
        parameters.orientation = "standard"
        result = labels._determine_image_dimensions(
            text=text, image_font=image_font, parameters=parameters
        )
        self.assertEqual((696, 141, 391, 72), result)

        parameters.orientation = "rotated"
        result = labels._determine_image_dimensions(
            text=text, image_font=image_font, parameters=parameters
        )
        self.assertEqual((461, 696, 391, 72), result)


class DetermineTextOffsetsTestCase(TestCase):
    def test_determine_text_offsets(self) -> None:
        parameters = labels.LabelParameters(
            configuration=self.example_configuration,
            font_family="Roboto",
            font_style="Medium",
        )

        # 1) Die cut/fixed size label.
        parameters.label_size = "62x29"
        parameters.orientation = "standard"
        result = labels._determine_text_offsets(
            height=271, width=696, text_height=72, text_width=391, parameters=parameters
        )
        self.assertEqual((152, 88), result)

        parameters.orientation = "rotated"
        result = labels._determine_text_offsets(
            width=271, height=696, text_height=72, text_width=391, parameters=parameters
        )
        self.assertEqual((0, 301), result)

        # 2) Endless label.
        parameters.label_size = "62"
        parameters.orientation = "standard"
        result = labels._determine_text_offsets(
            height=141, width=696, text_height=72, text_width=391, parameters=parameters
        )
        self.assertEqual((152, 24), result)

        parameters.orientation = "rotated"
        result = labels._determine_text_offsets(
            height=696, width=461, text_height=72, text_width=391, parameters=parameters
        )
        self.assertEqual((35, 301), result)


class CreateLabelImageTestCase(TestCase):
    def test_create_label_image(self) -> None:
        parameters = labels.LabelParameters(
            configuration=self.example_configuration,
            font_family="Roboto",
            font_style="Medium",
            text="Hello World!",
            label_size="62",
        )
        image = labels.create_label_image(parameters)
        self.addCleanup(image.close)
        reference = files("tests") / "data" / "hello_world.png"
        with as_file(reference) as path:
            with Image.open(path) as target_image:
                self.assertEqual(target_image.mode, image.mode)
                self.assertEqual(target_image.size, image.size)
                difference = ImageChops.difference(target_image, image)
                for index, pixel in enumerate(difference.getdata()):
                    self.assertEqual((0, 0, 0), pixel, index)


class ImageToPngBytesTestCase(TestCase):
    def test_image_to_png_bytes(self) -> None:
        reference = files("tests") / "data" / "hello_world.png"
        with as_file(reference) as path:
            with Image.open(path) as image:
                actual = labels.image_to_png_bytes(image)
            expected = path.read_bytes()
        self.assertEqual(expected, actual)


class GenerateLabelTestCase(TestCase):
    @mock.patch("brother_ql.raster.logger.warning")
    @mock.patch("brother_ql.conversion.logger.warning")
    def test_generate_label(self, _: mock.Mock, __: mock.Mock) -> None:
        parameters = labels.LabelParameters(
            configuration=self.example_configuration,
            font_family="Roboto",
            font_style="Medium",
            text="Hello World!",
        )

        # 1) Save image.
        with NamedTemporaryFile(suffix=".png") as save_to:
            result = labels.generate_label(
                parameters=parameters,
                configuration=parameters.configuration,
                save_image_to=save_to.name,
            )
            reference = files("tests") / "data" / "hello_world.png"
            with as_file(reference) as path:
                save_to.seek(0)
                self.assertEqual(path.read_bytes(), save_to.read())
        self.assertTrue(result.data)

        # 2) Endless label with standard orientation.
        parameters.label_size = "62"
        parameters.orientation = "standard"
        result = labels.generate_label(
            parameters=parameters, configuration=parameters.configuration
        )
        reference = (
            files("tests") / "data" / "hello_world__label_size_62__standard.data"
        )
        with as_file(reference) as path:
            self.assertEqual(path.read_bytes(), result.data)

        # 3) Endless label with rotated orientation.
        parameters.label_size = "62"
        parameters.orientation = "rotated"
        result = labels.generate_label(
            parameters=parameters, configuration=parameters.configuration
        )
        reference = files("tests") / "data" / "hello_world__label_size_62__rotated.data"
        with as_file(reference) as path:
            self.assertEqual(path.read_bytes(), result.data)

        # 4) Die cut label.
        for orientation in ["standard", "rotated"]:
            with self.subTest(orientation=orientation):
                parameters.label_size = "62x29"
                parameters.orientation = orientation
                result = labels.generate_label(
                    parameters=parameters,
                    configuration=parameters.configuration,
                )
                reference = (
                    files("tests")
                    / "data"
                    / f"hello_world__label_size_62x29__{orientation}.data"
                )
                with as_file(reference) as path:
                    self.assertEqual(path.read_bytes(), result.data)

        # 5) Red mode.
        parameters.label_size = "62red"
        parameters.orientation = "standard"
        parameters.configuration.printer.model = "QL-800"
        result = labels.generate_label(
            parameters=parameters, configuration=parameters.configuration
        )
        reference = (
            files("tests") / "data" / "hello_world__label_size_62red__standard.data"
        )
        with as_file(reference) as path:
            self.assertEqual(path.read_bytes(), result.data)


class PrintLabelTestCase(TestCase):
    def test_print_label(self) -> None:
        class Backend(BrotherQLBackendGeneric):
            def __init__(self, device_specifier: str) -> None:
                pass

        parameters = labels.LabelParameters(
            configuration=self.example_configuration,
        )
        qlr = BrotherQLRaster()
        qlr.data = b"My dummy data"

        # 1) One label.
        parameters.label_count = 1
        with mock.patch.object(labels.logger, "info") as info_mock, mock.patch.object(
            Backend, "write"
        ) as write_mock:
            labels.print_label(
                parameters=parameters,
                qlr=qlr,
                configuration=parameters.configuration,
                backend_class=Backend,
            )
        info_mock.assert_called_once_with("Printing label %d of %d ...", 1, 1)
        write_mock.assert_called_once_with(b"My dummy data")

        # 2) Multiple labels.
        parameters.label_count = 5
        with mock.patch.object(labels.logger, "info") as info_mock, mock.patch.object(
            Backend, "write"
        ) as write_mock:
            labels.print_label(
                parameters=parameters,
                qlr=qlr,
                configuration=parameters.configuration,
                backend_class=Backend,
            )
        info_mock.assert_has_calls(
            [mock.call("Printing label %d of %d ...", i, 5) for i in range(1, 6)],
            any_order=False,
        )
        self.assertEqual(5, info_mock.call_count, info_mock.call_args_list)
        write_mock.assert_has_calls([mock.call(b"My dummy data")] * 5)
        self.assertEqual(5, write_mock.call_count, write_mock.call_args_list)



from __future__ import annotations

try:
    from importlib.resources import as_file, files
except ImportError:
    from importlib_resources import as_file, files  # type: ignore[import-not-found,no-redef]  # noqa: E501
import base64
import os
import subprocess
from tempfile import NamedTemporaryFile
from threading import Thread
from time import sleep
from typing import cast
from unittest import mock

import requests
from brother_ql_web import web
from brother_ql_web.labels import LabelParameters

from tests import TestCase as _TestCase


class TestCase(_TestCase):
    process: subprocess.Popen[bytes] | None = None
    thread: Thread | None = None
    printer_file: str | None = None

    def setUp(self) -> None:
        super().setUp()
        self.process = None
        self.thread = None
        self.printer_file = None

    def tearDown(self) -> None:
        super().tearDown()
        if self.process:
            self.process.kill()
            if self.process.stdout:
                self.process.stdout.close()
            if self.process.stderr:
                self.process.stderr.close()
            self.process.wait()
        if self.thread:
            self.thread.join()
        if self.printer_file:
            os.unlink(self.printer_file)

    def run_server(self, log_level: str = "") -> None:
        self.printer_file = NamedTemporaryFile(delete=False).name

        def run() -> None:
            self.process = subprocess.Popen(
                [
                    "python",
                    "-m",
                    "brother_ql_web",
                    "--configuration",
                    self.example_configuration_path,
                    f"file://{self.printer_file}",
                ]
                + (["--log-level", log_level] if log_level else []),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )

        self.thread = Thread(target=run)
        self.thread.start()
        sleep(0.5)


class GetConfigTestCase(TestCase):
    pass


class IndexTestCase(TestCase):
    def test_index(self) -> None:
        self.run_server()
        response = requests.get("http://localhost:8013/", allow_redirects=False)
        self.assertEqual(303, response.status_code)
        self.assertEqual(
            "http://localhost:8013/labeldesigner", response.headers["Location"]
        )


class ServeStaticTestCase(TestCase):
    def test_serve_static(self) -> None:
        self.run_server()
        response = requests.get("http://localhost:8013/static/css/custom.css")
        self.assertEqual(200, response.status_code)
        reference = files("brother_ql_web") / "static" / "css" / "custom.css"
        with as_file(reference) as path:
            self.assertEqual(path.read_bytes(), response.content)


class LabeldesignerTestCase(TestCase):
    def test_labeldesigner(self) -> None:
        self.run_server()
        response = requests.get("http://localhost:8013/labeldesigner")
        self.assertEqual(200, response.status_code)
        self.assertIn(b"DejaVu Serif", response.content)


class GetLabelParametersTestCase(TestCase):
    def test_all_set(self) -> None:
        request = mock.Mock()
        request.params.decode.return_value = dict(
            font_family="DejaVu Serif (Book)",
            text="Hello World!",
            font_size="70",
            label_size="42",
            margin="13",
            threshold="50",
            align="left",
            orientation="rotated",
            margin_top="10",
            margin_bottom="20",
            margin_left="37",
            margin_right="38",
            label_count="20",
            high_quality="True",
        )
        request.app.config = {
            "brother_ql_web.configuration": self.example_configuration
        }

        parameters = web.get_label_parameters(request)
        self.assertEqual(
            LabelParameters(
                configuration=self.example_configuration,
                font_family="DejaVu Serif",
                font_style="Book",
                text="Hello World!",
                font_size=70,
                label_size="42",
                margin=13,
                threshold=50,
                align="left",
                orientation="rotated",
                margin_top=10,
                margin_bottom=20,
                margin_left=37,
                margin_right=38,
                label_count=20,
                high_quality=True,
            ),
            parameters,
        )

    def test_mostly_default_values(self) -> None:
        request = mock.Mock()
        request.params.decode.return_value = dict(font_family="Roboto (Regular)")
        request.app.config = {
            "brother_ql_web.configuration": self.example_configuration
        }

        parameters = web.get_label_parameters(request)
        self.assertEqual(
            LabelParameters(
                configuration=self.example_configuration,
                font_family="Roboto",
                font_style="Regular",
                text="",
                font_size=100,
                label_size="62",
                margin=10,
                threshold=70,
                align="center",
                orientation="standard",
                margin_top=24,
                margin_bottom=45,
                margin_left=35,
                margin_right=35,
                label_count=1,
                high_quality=False,
            ),
            parameters,
        )


class GetPreviewImageTestCase(TestCase):
    def test_base64(self) -> None:
        self.run_server()
        response = requests.get(
            "http://localhost:8013/api/preview/text?"
            "return_format=base64&text=Hello%20World!&"
            "label_size=62&font_family=Roboto%20(Medium)"
        )
        self.assertEqual(200, response.status_code)
        reference = files("tests") / "data" / "hello_world.png"
        with as_file(reference) as path:
            decoded = base64.b64decode(response.content)
            self.assertEqual(path.read_bytes(), decoded)

    def test_plain_bytes(self) -> None:
        self.run_server()
        response = requests.get(
            "http://localhost:8013/api/preview/text?"
            "return_format=png&text=Hello%20World!&"
            "label_size=62&font_family=Roboto%20(Medium)"
        )
        self.assertEqual(200, response.status_code)
        reference = files("tests") / "data" / "hello_world.png"
        with as_file(reference) as path:
            self.assertEqual(path.read_bytes(), response.content)

    def test_special_characters(self) -> None:
        self.run_server()
        # Minimal copy from the Firefox developer tools.
        headers = {
            "Content-Type": "multipart/form-data; boundary=---------------------------381934621323024354152349680295"  # noqa: E501
        }
        body = """
-----------------------------381934621323024354152349680295
Content-Disposition: form-data; name="text"

abcdefgö+ëŠ
-----------------------------381934621323024354152349680295
Content-Disposition: form-data; name="font_family"

Roboto (Medium)
-----------------------------381934621323024354152349680295
Content-Disposition: form-data; name="label_size"

62
-----------------------------381934621323024354152349680295--
""".encode()
        response = requests.post(
            "http://localhost:8013/api/preview/text?return_format=png",
            data=body,
            headers=headers,
        )
        self.assertEqual(200, response.status_code)
        self.assertLessEqual(14000, len(response.content))  # 14061 bytes


class PrintTextTestCase(TestCase):
    def test_error(self) -> None:
        self.run_server()
        response = requests.get("http://localhost:8013/api/print/text")
        self.assertEqual(200, response.status_code)
        self.assertEqual(
            (
                b'{"success": false, '
                b"\"error\": \"'NoneType' object has no attribute 'rpartition'\"}"
            ),
            response.content,
        )
        with open(cast(str, self.printer_file), mode="rb") as fd:
            self.assertEqual(b"", fd.read())

    def test_debug_mode(self) -> None:
        reference = files("tests") / "data" / "print_text__debug_mode.json"
        with as_file(reference) as path:
            expected = path.read_bytes()

        self.run_server(log_level="DEBUG")
        response = requests.get(
            "http://localhost:8013/api/print/text?"
            "text=Hello%20World!&label_size=62&"
            "font_family=Roboto%20(Medium)&orientation=standard"
        )
        self.assertEqual(200, response.status_code)
        self.assertEqual(expected, response.content)
        with open(cast(str, self.printer_file), mode="rb") as fd:
            self.assertEqual(b"", fd.read())

    def test_regular_mode(self) -> None:
        reference = (
            files("tests") / "data" / "hello_world__label_size_62__standard.data"
        )
        with as_file(reference) as path:
            expected = path.read_bytes()

        self.run_server()
        response = requests.get(
            "http://localhost:8013/api/print/text?"
            "text=Hello%20World!&label_size=62&"
            "font_family=Roboto%20(Medium)&orientation=standard"
        )
        self.assertEqual(200, response.status_code)
        self.assertEqual(b'{"success": true}', response.content)
        with open(cast(str, self.printer_file), mode="rb") as fd:
            self.assertEqual(expected, fd.read())


class MainTestCase(TestCase):
    pass



from __future__ import annotations

from unittest import TestCase

from brother_ql_web.font_helpers import get_fonts


class GetFontsTestCase(TestCase):
    # Reference: https://packages.ubuntu.com/lunar/all/fonts-roboto-unhinted/filelist
    ROBOTO_FILES = {
        "Roboto": {
            "Thin": "/usr/share/fonts/truetype/roboto/unhinted/RobotoTTF/Roboto-Thin.ttf",  # noqa: E501
            "Medium Italic": "/usr/share/fonts/truetype/roboto/unhinted/RobotoTTF/Roboto-MediumItalic.ttf",  # noqa: E501
            "Thin Italic": "/usr/share/fonts/truetype/roboto/unhinted/RobotoTTF/Roboto-ThinItalic.ttf",  # noqa: E501
            "Light Italic": "/usr/share/fonts/truetype/roboto/unhinted/RobotoTTF/Roboto-LightItalic.ttf",  # noqa: E501
            "Italic": "/usr/share/fonts/truetype/roboto/unhinted/RobotoTTF/Roboto-Italic.ttf",  # noqa: E501
            "Black Italic": "/usr/share/fonts/truetype/roboto/unhinted/RobotoTTF/Roboto-BlackItalic.ttf",  # noqa: E501
            "Medium": "/usr/share/fonts/truetype/roboto/unhinted/RobotoTTF/Roboto-Medium.ttf",  # noqa: E501
            "Bold": "/usr/share/fonts/truetype/roboto/unhinted/RobotoTTF/Roboto-Bold.ttf",  # noqa: E501
            "Black": "/usr/share/fonts/truetype/roboto/unhinted/RobotoTTF/Roboto-Black.ttf",  # noqa: E501
            "Light": "/usr/share/fonts/truetype/roboto/unhinted/RobotoTTF/Roboto-Light.ttf",  # noqa: E501
            "Regular": "/usr/share/fonts/truetype/roboto/unhinted/RobotoTTF/Roboto-Regular.ttf",  # noqa: E501
            "Bold Italic": "/usr/share/fonts/truetype/roboto/unhinted/RobotoTTF/Roboto-BoldItalic.ttf",  # noqa: E501
        },
        "Roboto Black": {
            "Italic": "/usr/share/fonts/truetype/roboto/unhinted/RobotoTTF/Roboto-BlackItalic.ttf",  # noqa: E501
            "Regular": "/usr/share/fonts/truetype/roboto/unhinted/RobotoTTF/Roboto-Black.ttf",  # noqa: E501
        },
        "Roboto Condensed": {
            "Italic": "/usr/share/fonts/truetype/roboto/unhinted/RobotoCondensed-Italic.ttf",  # noqa: E501
            "Bold": "/usr/share/fonts/truetype/roboto/unhinted/RobotoCondensed-Bold.ttf",  # noqa: E501
            "Medium": "/usr/share/fonts/truetype/roboto/unhinted/RobotoCondensed-Medium.ttf",  # noqa: E501
            "Bold Italic": "/usr/share/fonts/truetype/roboto/unhinted/RobotoCondensed-BoldItalic.ttf",  # noqa: E501
            "Light Italic": "/usr/share/fonts/truetype/roboto/unhinted/RobotoCondensed-LightItalic.ttf",  # noqa: E501
            "Light": "/usr/share/fonts/truetype/roboto/unhinted/RobotoCondensed-Light.ttf",  # noqa: E501
            "Medium Italic": "/usr/share/fonts/truetype/roboto/unhinted/RobotoCondensed-MediumItalic.ttf",  # noqa: E501
            "Regular": "/usr/share/fonts/truetype/roboto/unhinted/RobotoCondensed-Regular.ttf",  # noqa: E501
        },
        "Roboto Condensed Light": {
            "Italic": "/usr/share/fonts/truetype/roboto/unhinted/RobotoCondensed-LightItalic.ttf",  # noqa: E501
            "Regular": "/usr/share/fonts/truetype/roboto/unhinted/RobotoCondensed-Light.ttf",  # noqa: E501
        },
        "Roboto Condensed Medium": {
            "Regular": "/usr/share/fonts/truetype/roboto/unhinted/RobotoCondensed-Medium.ttf",  # noqa: E501
            "Italic": "/usr/share/fonts/truetype/roboto/unhinted/RobotoCondensed-MediumItalic.ttf",  # noqa: E501
        },
        "Roboto Light": {
            "Italic": "/usr/share/fonts/truetype/roboto/unhinted/RobotoTTF/Roboto-LightItalic.ttf",  # noqa: E501
            "Regular": "/usr/share/fonts/truetype/roboto/unhinted/RobotoTTF/Roboto-Light.ttf",  # noqa: E501
        },
        "Roboto Medium": {
            "Italic": "/usr/share/fonts/truetype/roboto/unhinted/RobotoTTF/Roboto-MediumItalic.ttf",  # noqa: E501
            "Regular": "/usr/share/fonts/truetype/roboto/unhinted/RobotoTTF/Roboto-Medium.ttf",  # noqa: E501
        },
        "Roboto Thin": {
            "Regular": "/usr/share/fonts/truetype/roboto/unhinted/RobotoTTF/Roboto-Thin.ttf",  # noqa: E501
            "Italic": "/usr/share/fonts/truetype/roboto/unhinted/RobotoTTF/Roboto-ThinItalic.ttf",  # noqa: E501
        },
    }

    def assert_font_dictionary_subset(
        self,
        expected: dict[str, dict[str, str]],
        actual: dict[str, dict[str, str]],
        not_in_name: str = "",
    ) -> None:
        for font_name, font_styles in expected.items():
            self.assertIn(font_name, actual)
            actual_styles = actual[font_name]
            with self.subTest(font_name=font_name):
                for font_style, path in font_styles.items():
                    self.assertIn(font_style, actual_styles)
                    self.assertEqual(path, actual_styles[font_style], font_style)
            if not_in_name:
                self.assertNotIn(not_in_name, font_name)

    def test_get_all(self) -> None:
        fonts = get_fonts()
        self.assert_font_dictionary_subset(expected=self.ROBOTO_FILES, actual=fonts)

    def test_get_from_folder(self) -> None:
        fonts = get_fonts(folder="/usr/share/fonts/truetype/roboto/unhinted/RobotoTTF")
        expected = {
            key: value
            for key, value in self.ROBOTO_FILES.items()
            if "Condensed" not in key
        }
        self.assert_font_dictionary_subset(
            expected=expected, actual=fonts, not_in_name="Condensed"
        )



from __future__ import annotations

from unittest import mock

from brother_ql_web import utils

from tests import TestCase


class CollectFontsTestCase(TestCase):
    def test_without_folder(self) -> None:
        dummy_fonts = {
            "DejaVu Serif": {"Book": "dummy", "Regular": "dummy2"},
            "Font": {"Style": "path"},
        }
        configuration = self.example_configuration
        with mock.patch.object(
            utils, "get_fonts", return_value=dummy_fonts
        ) as get_mock:
            fonts = utils.collect_fonts(configuration)
        get_mock.assert_called_once_with()
        self.assertEqual(dummy_fonts, fonts)

    def test_with_folder(self) -> None:
        dummy_fonts = {
            "DejaVu Serif": {"Book": "dummy", "Regular": "dummy2"},
            "Font": {"Style": "path"},
        }
        folder_fonts = {
            "Font": {"Style": "/path", "Another Style": "/path2"},
            "Symbol": {"Regular": "/path3"},
        }

        def get_fonts(directory: None | str = None) -> dict[str, dict[str, str]]:
            return folder_fonts if directory else dummy_fonts

        configuration = self.example_configuration
        configuration.server.additional_font_folder = "/another_path"
        with mock.patch.object(utils, "get_fonts", side_effect=get_fonts) as get_mock:
            fonts = utils.collect_fonts(configuration)

        get_mock.assert_has_calls(
            [mock.call(), mock.call("/another_path")], any_order=False
        )
        self.assertEqual(2, get_mock.call_count, get_mock.call_args_list)

        self.assertEqual(
            {
                "DejaVu Serif": {"Book": "dummy", "Regular": "dummy2"},
                "Font": {"Style": "/path", "Another Style": "/path2"},
                "Symbol": {"Regular": "/path3"},
            },
            fonts,
        )


class GetLabelSizesTestCase(TestCase):
    def test_get_label_sizes(self) -> None:
        sizes = utils.get_label_sizes()
        self.assertIn(("38", "38mm endless"), sizes)
        self.assertIn(("62x29", "62mm x 29mm die-cut"), sizes)


class GetBackendClassTestCase(TestCase):
    def test_unknown_backend(self) -> None:
        configuration = self.example_configuration
        configuration.printer.printer = "dummy"

        with self.assertRaisesRegex(
            utils.BackendGuessingError,
            r"^Couln't guess the backend to use from the printer string descriptor$",
        ):
            utils.get_backend_class(configuration)

    def test_known_backend(self) -> None:
        from brother_ql.backends.linux_kernel import BrotherQLBackendLinuxKernel

        configuration = self.example_configuration
        backend = utils.get_backend_class(configuration)
        self.assertIs(backend, BrotherQLBackendLinuxKernel)



from __future__ import annotations

import sys
from typing import cast

from brother_ql.backends import backend_factory, BrotherQLBackendGeneric, guess_backend
from brother_ql.devicedependent import label_type_specs, label_sizes
from brother_ql_web.configuration import Configuration
from brother_ql_web.font_helpers import get_fonts

if sys.version_info < (3, 9):
    from typing import Type

    BACKEND_TYPE = Type[BrotherQLBackendGeneric]
else:
    BACKEND_TYPE = type[BrotherQLBackendGeneric]


def collect_fonts(configuration: Configuration) -> dict[str, dict[str, str]]:
    fonts = get_fonts()
    if configuration.server.additional_font_folder:
        fonts.update(get_fonts(configuration.server.additional_font_folder))
    return fonts


def get_label_sizes() -> list[tuple[str, str]]:
    return [(name, cast(str, label_type_specs[name]["name"])) for name in label_sizes]


class BackendGuessingError(ValueError):
    pass


def get_backend_class(configuration: Configuration) -> BACKEND_TYPE:
    try:
        selected_backend = guess_backend(configuration.printer.printer)
    except ValueError:
        raise BackendGuessingError(
            "Couln't guess the backend to use from the printer string descriptor"
        )
    return cast(
        BACKEND_TYPE,
        backend_factory(selected_backend)["backend_class"],
    )



from __future__ import annotations

import logging
from typing import Any


def patch_deprecation_warning() -> None:
    """
    Avoid the deprecation warning from `brother_ql.devicedependent`. This has been
    fixed in the Git version, but not in the PyPI version:
    https://github.com/pklaus/brother_ql/commit/5c2b72b18bcf436c116f180a9147cbb6805958f5
    """
    original_logger = logging.getLogger("brother_ql.devicedependent").warning

    def warn(message: str, *args: Any, **kwargs: Any) -> None:
        if (
            message
            == "deprecation warning: brother_ql.devicedependent is deprecated and will be removed in a future release"  # noqa: E501
        ):
            return
        original_logger(message, *args, **kwargs)

    logging.getLogger("brother_ql.devicedependent").warn = warn  # type: ignore[assignment,method-assign]  # noqa: E501


patch_deprecation_warning()


import brother_ql.conversion  # noqa: E402
import PIL  # noqa: E402


# Renamed in version 2.7.0:
# https://pillow.readthedocs.io/en/stable/releasenotes/2.7.0.html#antialias-renamed-to-lanczos
brother_ql.conversion.Image.ANTIALIAS = PIL.Image.LANCZOS  # type: ignore[attr-defined]


__all__: list[str] = []



def main() -> None:
    import logging

    from brother_ql_web import cli, utils, web
    from brother_ql_web.configuration import Configuration

    cli_parameters = cli.get_parameters()
    configuration = Configuration.from_json(cli_parameters.configuration)
    cli.update_configuration_from_parameters(
        configuration=configuration, parameters=cli_parameters
    )
    logging.basicConfig(level=configuration.server.log_level)
    web.main(
        configuration=configuration,
        fonts=utils.collect_fonts(configuration),
        label_sizes=utils.get_label_sizes(),
        backend_class=utils.get_backend_class(configuration),
    )


if __name__ == "__main__":
    main()



from __future__ import annotations

import json
from tempfile import NamedTemporaryFile
from typing import cast, Dict, Union

from brother_ql_web.configuration import (
    Configuration,
    Font,
    LabelConfiguration,
    PrinterConfiguration,
    ServerConfiguration,
    WebsiteConfiguration,
)

from tests import TestCase


CUSTOM_CONFIGURATION = """
{
  "server": {
    "port": 1337,
    "host": "test.local",
    "log_level": "ERROR",
    "additional_font_folder": ""
  },
  "printer": {
    "model": "QL-800",
    "printer": "file:///dev/usb/lp1"
  },
  "label": {
    "default_size": "62",
    "default_orientation": "standard",
    "default_font_size": 42,
    "default_fonts": [
      {
        "family": "Minion Pro",
        "style": "Semibold"
      },
      {
        "family": "Linux Libertine",
        "style": "Regular"
      },
      {
        "family": "DejaVu Serif",
        "style": "Book"
      }
    ],
    "default_font": null
  },
  "website": {
    "html_title": "Label Designer",
    "page_title": "Brother QL Label Designer",
    "page_headline": "Design your label and print it!"
  }
}
"""


class ConfigurationTestCase(TestCase):
    @property
    def example_json(self) -> dict[str, dict[str, str | int | dict[str, str]]]:
        with open(self.example_configuration_path) as fd:
            return cast(
                Dict[str, Dict[str, Union[str, int, Dict[str, str]]]], json.load(fd)
            )

    def test_from_json(self) -> None:
        with NamedTemporaryFile(suffix=".json", mode="w+t") as json_file:
            json_file.write(CUSTOM_CONFIGURATION)
            json_file.seek(0)

            configuration = Configuration.from_json(json_file.name)
            self.assertEqual(
                ServerConfiguration(
                    port=1337,
                    host="test.local",
                    log_level="ERROR",
                    additional_font_folder="",
                ),
                configuration.server,
            )
            self.assertEqual(
                PrinterConfiguration(model="QL-800", printer="file:///dev/usb/lp1"),
                configuration.printer,
            )
            self.assertEqual(
                LabelConfiguration(
                    default_size="62",
                    default_orientation="standard",
                    default_font_size=42,
                    default_fonts=[
                        Font(family="Minion Pro", style="Semibold"),
                        Font(family="Linux Libertine", style="Regular"),
                        Font(family="DejaVu Serif", style="Book"),
                    ],
                    default_font=None,
                ),
                configuration.label,
            )
            self.assertEqual(
                WebsiteConfiguration(
                    html_title="Label Designer",
                    page_title="Brother QL Label Designer",
                    page_headline="Design your label and print it!",
                ),
                configuration.website,
            )

    def test_from_json__too_many_keys(self) -> None:
        with NamedTemporaryFile(suffix=".json", mode="w+t") as json_file:
            data = self.example_json
            data["key"] = {"value": 42}
            json_file.write(json.dumps(data))
            json_file.seek(0)

            with self.assertRaisesRegex(
                ValueError,
                r"^Unknown configuration values: \{'key': \{'value': 42\}\}$",
            ):
                Configuration.from_json(json_file.name)

    def test_from_json__missing_server_key(self) -> None:
        with NamedTemporaryFile(suffix=".json", mode="w+t") as json_file:
            data = self.example_json
            del data["server"]
            json_file.write(json.dumps(data))
            json_file.seek(0)

            configuration = Configuration.from_json(json_file.name)
            self.assertEqual(ServerConfiguration(), configuration.server)

    def test_from_json__missing_printer_key(self) -> None:
        with NamedTemporaryFile(suffix=".json", mode="w+t") as json_file:
            data = self.example_json
            del data["printer"]
            json_file.write(json.dumps(data))
            json_file.seek(0)

            with self.assertRaisesRegex(ValueError, r"^Printer configuration missing$"):
                Configuration.from_json(json_file.name)

    def test_to_json(self) -> None:
        with NamedTemporaryFile(suffix=".json", mode="w+t") as json_file:
            json_file.write(CUSTOM_CONFIGURATION)
            json_file.seek(0)

            configuration = Configuration.from_json(json_file.name)
            self.assertEqual(
                CUSTOM_CONFIGURATION.strip(), configuration.to_json().strip()
            )


class ServerConfigurationTestCase(TestCase):
    def test_is_in_debug_mode(self) -> None:
        self.assertTrue(ServerConfiguration(log_level="DEBUG").is_in_debug_mode)

        for level in ["INFO", "WARNING", "ERROR"]:
            with self.subTest(level=level):
                self.assertFalse(ServerConfiguration(log_level=level).is_in_debug_mode)


class PrinterConfigurationTestCase(TestCase):
    pass


class FontTestCase(TestCase):
    pass


class LabelConfigurationTestCase(TestCase):
    def test_post_init(self) -> None:
        default_fonts: list[Font | dict[str, str]] = [
            Font(family="Font family", style="Regular"),
            {"family": "Another font family", "style": "Bold"},
        ]
        configuration = LabelConfiguration(default_fonts=default_fonts)  # type: ignore[arg-type]  # noqa: E501
        self.assertEqual(
            [
                Font(family="Font family", style="Regular"),
                Font(family="Another font family", style="Bold"),
            ],
            configuration.default_fonts,
        )


class WebsiteConfigurationTestCase(TestCase):
    pass



from unittest import mock

from brother_ql_web.__main__ import main
from brother_ql_web.configuration import Font

from tests import TestCase


class Backend:
    pass


class MainTestCase(TestCase):
    def test_main(self) -> None:
        fonts = {
            "DejaVu Serif": {"Book": "dummy", "Regular": "dummy2"},
        }
        label_sizes = [("62", "62mm labels"), "42", "42mm labels"]

        with mock.patch(
            "sys.argv",
            [
                "brother_ql_web/__main__",
                "--configuration",
                self.example_configuration_path,
            ],
        ), mock.patch("logging.basicConfig") as basic_config_mock, mock.patch(
            "brother_ql_web.utils.collect_fonts", return_value=fonts
        ) as fonts_mock, mock.patch(
            "brother_ql_web.utils.get_label_sizes", return_value=label_sizes
        ) as labels_mock, mock.patch(
            "brother_ql_web.utils.get_backend_class", return_value=Backend
        ) as backend_mock, mock.patch(
            "brother_ql_web.web.main"
        ) as main_mock:
            main()

        configuration = self.example_configuration
        configuration.label.default_font = Font(family="DejaVu Serif", style="Book")

        basic_config_mock.assert_called_once_with(level="WARNING")
        main_mock.assert_called_once_with(
            configuration=configuration,
            fonts=fonts,
            label_sizes=label_sizes,
            backend_class=Backend,
        )
        fonts_mock.assert_called_once_with(configuration)
        labels_mock.assert_called_once_with()
        backend_mock.assert_called_once_with(configuration)



from __future__ import annotations

import logging
import subprocess
from collections import defaultdict


logger = logging.getLogger(__name__)
del logging


def get_fonts(folder: str | None = None) -> dict[str, dict[str, str]]:
    """
    Scan a folder (or the system) for .ttf / .otf fonts and
    return a dictionary of the structure  family -> style -> file path
    """
    fonts: dict[str, dict[str, str]] = defaultdict(dict)
    if folder:
        cmd = ["fc-scan", "--format", "%{file}:%{family}:style=%{style}\n", folder]
    else:
        cmd = ["fc-list", ":", "file", "family", "style"]
    for line in subprocess.check_output(cmd).decode("utf-8").split("\n"):
        line = line.strip()
        if not line:
            continue
        if "otf" not in line and "ttf" not in line:
            continue
        parts = line.split(":")
        if "style=" not in line or len(parts) < 3:
            # fc-list did not output all desired properties
            logger.warning("skipping invalid font %s", line)
            continue
        path = parts[0]
        families = parts[1].strip().split(",")
        styles = parts[2].split("=")[1].split(",")
        if len(families) == 1 and len(styles) > 1:
            families = [families[0]] * len(styles)
        elif len(families) > 1 and len(styles) == 1:
            styles = [styles[0]] * len(families)
        if len(families) != len(styles):
            logger.debug("Problem with this font: %s", line)
            continue
        for i in range(len(families)):
            fonts[families[i]][styles[i]] = path
            # logger.debug("Added this font: %s", (families[i], styles[i], path))
    return dict(fonts)



from __future__ import annotations

from argparse import Namespace
from io import StringIO
from contextlib import redirect_stderr
from unittest import mock

from tests import TestCase  # Silence useless deprecation warning.

from brother_ql_web import cli
from brother_ql_web.configuration import Configuration, Font


class LogLevelTypeTestCase(TestCase):
    def test_valid(self) -> None:
        self.assertEqual(40, cli.log_level_type("ERROR"))
        self.assertEqual(40, cli.log_level_type("error"))

    def test_invalid(self) -> None:
        with self.assertRaisesRegex(
            AttributeError, "^module 'logging' has no attribute 'XYZ'$"
        ):
            cli.log_level_type("xyz")


class GetParametersTestCase(TestCase):
    def test_get_parameters(self) -> None:
        with mock.patch(
            "sys.argv",
            [
                "brother_ql_web/__main__.py",
                "--configuration",
                self.example_configuration_path,
                "--log-level",
                "INFO",
                "/dev/printer1",
            ],
        ):
            parameters = cli.get_parameters()

        self.assertEqual(
            Namespace(
                port=False,
                log_level=20,
                font_folder=False,
                default_label_size=False,
                default_orientation=False,
                model=False,
                printer="/dev/printer1",
                configuration=self.example_configuration_path,
            ),
            parameters,
        )


class ChooseDefaultFontTestCase(TestCase):
    FONTS = {
        "family1": {
            "Regular": "family1/regular.otf",
            "Italic": "family1/italic.otf",
        },
        "family2": {
            "Regular": "family2/regular.ttf",
            "Bold": "family2/bold.ttf",
        },
        "family 3": {"Italic": "family3/italic.otf"},
    }

    def assert_is_valid_font(self, font: Font | None) -> None:
        self.assertIsNotNone(font)
        self.assertIn(font.family, self.FONTS)  # type: ignore[union-attr]
        self.assertIn(font.style, self.FONTS[font.family])  # type: ignore[union-attr]

    def test_invalid_only(self) -> None:
        configuration = self.example_configuration
        invalid_font = Font(family="invalid", style="Regular")
        configuration.label.default_fonts = [invalid_font]

        stderr = StringIO()
        with redirect_stderr(stderr):
            self.assertIsNone(configuration.label.default_font)
            cli._choose_default_font(fonts=self.FONTS, configuration=configuration)
            self.assertIsNotNone(configuration.label.default_font)
            font = configuration.label.default_font
            self.assertNotEqual(invalid_font, font)
            self.assert_is_valid_font(font)

        self.assertEqual(
            (
                "Could not find any of the default fonts. Choosing a random one.\n"
                f"The default font is now set to: {font}\n"
            ),
            stderr.getvalue(),
        )

    def test_no_font_given(self) -> None:
        configuration = self.example_configuration
        configuration.label.default_fonts = []

        # Retrieving only two fonts proved to be rather unstable, as we would often
        # choose the same font both times, thus the test would not be representative.
        # By doing ten iterations and assuring that at least three different fonts have
        # been retrieved, we test the same, but in a more stable fashion. Evaluations
        # have shown the we usually get 3-5 different fonts out of ten samples.
        stderr = StringIO()
        fonts = []
        with redirect_stderr(stderr):
            self.assertIsNone(configuration.label.default_font)

            for _ in range(10):
                cli._choose_default_font(fonts=self.FONTS, configuration=configuration)
                self.assertIsNotNone(configuration.label.default_font)
                font = configuration.label.default_font
                self.assert_is_valid_font(font)
                fonts.append(font)

        self.assertGreaterEqual(len(set(fonts)), 3, fonts)

        self.assertEqual(
            "".join(
                (
                    "Could not find any of the default fonts. Choosing a random one.\n"
                    f"The default font is now set to: {font}\n"
                )
                for font in fonts
            ),
            stderr.getvalue(),
        )

    def test_first_valid_font_chosen(self) -> None:
        configuration = self.example_configuration
        configuration.label.default_fonts = [
            Font(family="invalid", style="Regular"),
            Font(family="family2", style="Bold"),
            Font(family="family1", style="Regular"),
        ]

        stderr = StringIO()
        with redirect_stderr(stderr):
            self.assertIsNone(configuration.label.default_font)
            cli._choose_default_font(fonts=self.FONTS, configuration=configuration)
            self.assertIsNotNone(configuration.label.default_font)
            font = configuration.label.default_font
            self.assert_is_valid_font(font)
            self.assertEqual(Font(family="family2", style="Bold"), font)

        self.assertEqual("", stderr.getvalue())


class UpdateConfigurationFromParametersTestCase(TestCase):
    @staticmethod
    def dummy_choose_default_font(
        fonts: dict[str, dict[str, str]], configuration: Configuration
    ) -> Configuration:
        configuration.label.default_font = Font(family="My Family", style="Bold")
        return configuration

    @property
    def expected_configuration(self) -> Configuration:
        configuration = self.example_configuration
        configuration.label.default_font = Font(family="My Family", style="Bold")
        return configuration

    def setUp(self) -> None:
        super().setUp()

        choose_patcher = mock.patch.object(
            cli, "_choose_default_font", side_effect=self.dummy_choose_default_font
        )
        self.choose_mock = choose_patcher.start()
        self.addCleanup(choose_patcher.stop)

    def test_no_overwrites(self) -> None:
        configuration = self.example_configuration
        parameters = Namespace(
            port=False,
            log_level=False,
            font_folder=False,
            default_label_size=False,
            default_orientation=False,
            model=False,
            printer=False,
            configuration=self.example_configuration_path,
        )

        cli.update_configuration_from_parameters(
            parameters=parameters, configuration=configuration
        )
        self.choose_mock.assert_called_once()

        self.assertEqual(self.expected_configuration, configuration)

    def test_invalid_label_size(self) -> None:
        configuration = self.example_configuration
        parameters = Namespace(
            port=False,
            log_level=False,
            font_folder=False,
            default_label_size=False,
            default_orientation=False,
            model=False,
            printer=False,
            configuration=self.example_configuration_path,
        )
        configuration.label.default_size = "dummy"

        # Do not assert the label sizes, but make sure that the value list starts with
        # a number.
        with self.assertRaisesRegex(
            cli.InvalidLabelSize,
            r"^Invalid default label size\. Please choose one of the following:\n\d.+$",  # noqa: E501
        ):
            cli.update_configuration_from_parameters(
                parameters=parameters, configuration=configuration
            )

    def test_no_fonts_found(self) -> None:
        configuration = self.example_configuration
        parameters = Namespace(
            port=False,
            log_level=False,
            font_folder=False,
            default_label_size=False,
            default_orientation=False,
            model=False,
            printer=False,
            configuration=self.example_configuration_path,
        )

        with mock.patch.object(cli, "collect_fonts", return_value=[]) as collect_mock:
            with self.assertRaisesRegex(
                cli.NoFontFound,
                '^Not a single font was found on your system. Please install some or use the "--font-folder" argument.$',  # noqa: E501
            ):
                cli.update_configuration_from_parameters(
                    parameters=parameters, configuration=configuration
                )
        collect_mock.assert_called_once_with(configuration)

    def test_overwrite_port(self) -> None:
        configuration = self.example_configuration
        parameters = Namespace(
            port=1337,
            log_level=False,
            font_folder=False,
            default_label_size=False,
            default_orientation=False,
            model=False,
            printer=False,
            configuration=self.example_configuration_path,
        )

        cli.update_configuration_from_parameters(
            parameters=parameters, configuration=configuration
        )
        self.choose_mock.assert_called_once()

        expected_configuration = self.expected_configuration
        expected_configuration.server.port = 1337
        self.assertEqual(expected_configuration, configuration)

    def test_overwrite_log_level(self) -> None:
        for log_level in [20, "INFO"]:
            with self.subTest(log_level=log_level):
                self.choose_mock.reset_mock()

                configuration = self.example_configuration
                parameters = Namespace(
                    port=False,
                    log_level=log_level,
                    font_folder=False,
                    default_label_size=False,
                    default_orientation=False,
                    model=False,
                    printer=False,
                    configuration=self.example_configuration_path,
                )

                cli.update_configuration_from_parameters(
                    parameters=parameters, configuration=configuration
                )
                self.choose_mock.assert_called_once()

                expected_configuration = self.expected_configuration
                expected_configuration.server.log_level = "INFO"
                self.assertEqual(expected_configuration, configuration)

    def test_overwrite_font_folder(self) -> None:
        configuration = self.example_configuration
        parameters = Namespace(
            port=False,
            log_level=False,
            font_folder="/path/to/fonts",
            default_label_size=False,
            default_orientation=False,
            model=False,
            printer=False,
            configuration=self.example_configuration_path,
        )

        dummy_fonts = {"DejaVu Serif": {"Book": "dummy"}}
        with mock.patch(
            "brother_ql_web.utils.get_fonts", return_value=dummy_fonts
        ) as fonts_mock:
            cli.update_configuration_from_parameters(
                parameters=parameters, configuration=configuration
            )
        self.choose_mock.assert_called_once()
        fonts_mock.assert_has_calls(
            [mock.call(), mock.call("/path/to/fonts")],
            any_order=False,
        )
        self.assertEqual(2, fonts_mock.call_count, fonts_mock.call_args_list)

        expected_configuration = self.expected_configuration
        expected_configuration.server.additional_font_folder = "/path/to/fonts"
        self.assertEqual(expected_configuration, configuration)

    def test_overwrite_printer(self) -> None:
        configuration = self.example_configuration
        parameters = Namespace(
            port=False,
            log_level=False,
            font_folder=False,
            default_label_size=False,
            default_orientation=False,
            model=False,
            printer="/dev/printer42",
            configuration=self.example_configuration_path,
        )

        cli.update_configuration_from_parameters(
            parameters=parameters, configuration=configuration
        )
        self.choose_mock.assert_called_once()

        expected_configuration = self.expected_configuration
        expected_configuration.printer.printer = "/dev/printer42"
        self.assertEqual(expected_configuration, configuration)

    def test_overwrite_model(self) -> None:
        configuration = self.example_configuration
        parameters = Namespace(
            port=False,
            log_level=False,
            font_folder=False,
            default_label_size=False,
            default_orientation=False,
            model="QL-800",
            printer=False,
            configuration=self.example_configuration_path,
        )

        cli.update_configuration_from_parameters(
            parameters=parameters, configuration=configuration
        )
        self.choose_mock.assert_called_once()

        expected_configuration = self.expected_configuration
        expected_configuration.printer.model = "QL-800"
        self.assertEqual(expected_configuration, configuration)

    def test_overwrite_default_label_size(self) -> None:
        configuration = self.example_configuration
        parameters = Namespace(
            port=False,
            log_level=False,
            font_folder=False,
            default_label_size="38",
            default_orientation=False,
            model=False,
            printer=False,
            configuration=self.example_configuration_path,
        )

        cli.update_configuration_from_parameters(
            parameters=parameters, configuration=configuration
        )
        self.choose_mock.assert_called_once()

        expected_configuration = self.expected_configuration
        expected_configuration.label.default_size = "38"
        self.assertEqual(expected_configuration, configuration)

    def test_overwrite_default_orientation(self) -> None:
        configuration = self.example_configuration
        parameters = Namespace(
            port=False,
            log_level=False,
            font_folder=False,
            default_label_size=False,
            default_orientation="my_orientation",
            model=False,
            printer=False,
            configuration=self.example_configuration_path,
        )

        cli.update_configuration_from_parameters(
            parameters=parameters, configuration=configuration
        )
        self.choose_mock.assert_called_once()

        expected_configuration = self.expected_configuration
        expected_configuration.label.default_orientation = "my_orientation"
        self.assertEqual(expected_configuration, configuration)



from setuptools import setup, find_namespace_packages
from pathlib import Path


ROOT_DIRECTORY = Path(__file__).parent.resolve()


setup(
    name="brother_ql_web",
    description="Simple label designer API and web API for Brother QL label printers",
    version="0.2.1",
    license="GPL-3.0-only",
    long_description=(ROOT_DIRECTORY / "README.md").read_text(encoding="utf-8"),
    long_description_content_type="text/markdown",
    author="Philipp Klaus (initial), FriedrichFröbel",
    url="https://github.com/FriedrichFroebel/brother_ql_web/",
    packages=find_namespace_packages(include=["brother_ql_web", "brother_ql_web.*"]),
    include_package_data=True,
    python_requires=">=3.8, <4",
    install_requires=[
        "brother_ql",
        "bottle",
        "jinja2",
        "Pillow>=8",
    ],
    extras_require={
        "dev": [
            "black",
            "flake8",
            "pep8-naming",
            "flake8-bugbear",
            "requests",
            'importlib-resources; python_version < "3.9"',
        ],
        "mypy": [
            "mypy",
            "types-Pillow",
            "types-requests",
        ],
    },
    entry_points={
        "console_scripts": [
            "brother_ql_web=brother_ql_web.__main__:main",
        ],
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Framework :: Bottle",
        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Multimedia :: Graphics",
    ],
)



from __future__ import annotations

import logging
import random
import sys
from argparse import ArgumentParser, Namespace
from typing import cast

from brother_ql.devicedependent import models, label_sizes
from brother_ql_web.configuration import Configuration, Font
from brother_ql_web.utils import collect_fonts


logger = logging.getLogger(__name__)


def log_level_type(value: str) -> int:
    return cast(int, getattr(logging, value.upper()))


def get_parameters() -> Namespace:
    parser = ArgumentParser(
        description=(
            "This is a web service to print labels on Brother QL label printers."
        )
    )
    parser.add_argument("--port", default=False)
    parser.add_argument("--log-level", type=log_level_type, default=False)
    parser.add_argument(
        "--font-folder", default=False, help="folder for additional .ttf/.otf fonts"
    )
    parser.add_argument(
        "--default-label-size",
        default=False,
        help="Label size inserted in your printer. Defaults to 62.",
    )
    parser.add_argument(
        "--default-orientation",
        default=False,
        choices=("standard", "rotated"),
        help=(
            'Label orientation, defaults to "standard". '
            'To turn your text by 90°, state "rotated".'
        ),
    )
    parser.add_argument(
        "--model",
        default=False,
        choices=models,
        help="The model of your printer (default: QL-500)",
    )
    parser.add_argument(
        "printer",
        nargs="?",
        default=False,
        help=(
            "String descriptor for the printer to use "
            "(like tcp://192.168.0.23:9100 or file:///dev/usb/lp0)"
        ),
    )
    parser.add_argument(
        "--configuration",
        required=True,
        type=str,
        help="Path to the configuration file to get the basic values from.",
    )

    return parser.parse_args()


class InvalidLabelSize(ValueError):
    pass


class NoFontFound(SystemError):
    pass


def _choose_default_font(
    fonts: dict[str, dict[str, str]], configuration: Configuration
) -> None:
    for font in configuration.label.default_fonts:
        try:
            fonts[font.family][font.style]
            configuration.label.default_font = font
            logger.debug("Selected the following default font: %s", font)
            break
        except KeyError:
            pass
    else:
        sys.stderr.write(
            "Could not find any of the default fonts. Choosing a random one.\n"
        )
        family = random.choice(list(fonts.keys()))
        style = random.choice(list(fonts[family].keys()))
        configuration.label.default_font = Font(family=family, style=style)
        sys.stderr.write(
            f"The default font is now set to: {configuration.label.default_font}\n"
        )


def update_configuration_from_parameters(
    parameters: Namespace, configuration: Configuration
) -> None:
    # Server configuration.
    if parameters.port:
        configuration.server.port = parameters.port
    if parameters.log_level:
        # `log_level` will be numeric if parsed from argv, so we enforce the name here.
        level = parameters.log_level
        configuration.server.log_level = (
            logging.getLevelName(level) if isinstance(level, int) else level
        )
    if parameters.font_folder:
        configuration.server.additional_font_folder = parameters.font_folder

    # Printer configuration.
    if parameters.printer:
        configuration.printer.printer = parameters.printer
    if parameters.model:
        configuration.printer.model = parameters.model

    # Label configuration.
    if parameters.default_label_size:
        configuration.label.default_size = parameters.default_label_size
    if parameters.default_orientation:
        configuration.label.default_orientation = parameters.default_orientation

    # Configuration issues.
    if configuration.label.default_size not in label_sizes:
        raise InvalidLabelSize(
            "Invalid default label size. Please choose one of the following:\n"
            + " ".join(label_sizes)
        )

    # System issues.
    fonts = collect_fonts(configuration)
    if not fonts:
        raise NoFontFound(
            "Not a single font was found on your system. "
            'Please install some or use the "--font-folder" argument.\n'
        )

    # Set font data.
    _choose_default_font(fonts=fonts, configuration=configuration)



from __future__ import annotations

import json
from dataclasses import dataclass, field as dataclass_field, fields as dataclass_fields
from typing import Any, cast


@dataclass
class Configuration:
    server: ServerConfiguration
    printer: PrinterConfiguration
    label: LabelConfiguration
    website: WebsiteConfiguration

    @classmethod
    def from_json(cls, json_file: str) -> Configuration:
        with open(json_file, mode="r") as fd:
            parsed: dict[str, Any] = json.load(fd)
        kwargs: dict[str, Any] = {}
        global_variables = globals()
        for field in dataclass_fields(cls):
            name: str = field.name
            field_type: str = cast(str, field.type)
            field_class = global_variables[field_type]
            kwargs_inner = parsed.pop(name, None)
            if name == "printer" and not kwargs_inner:
                raise ValueError("Printer configuration missing")
            if not kwargs_inner:
                instance = field_class()
            else:
                instance = field_class(**kwargs_inner)
            kwargs[name] = instance
        if parsed:
            raise ValueError(f"Unknown configuration values: {parsed}")
        return cls(**kwargs)

    def to_json(self) -> str:
        return json.dumps(self, indent=2, default=lambda o: o.__dict__)


@dataclass
class ServerConfiguration:
    port: int = 8013
    host: str = ""
    log_level: str = "WARNING"
    additional_font_folder: str = ""

    @property
    def is_in_debug_mode(self) -> bool:
        return self.log_level == "DEBUG"


@dataclass
class PrinterConfiguration:
    model: str
    printer: str


@dataclass(frozen=True)
class Font:
    family: str
    style: str


@dataclass
class LabelConfiguration:
    default_size: str = "62"
    default_orientation: str = "standard"
    default_font_size: int = 70
    default_fonts: list[Font] = dataclass_field(default_factory=list)
    default_font: Font | None = None

    def __post_init__(self) -> None:
        self.default_fonts = [
            font if isinstance(font, Font) else Font(**font)  # type: ignore[arg-type]
            for font in self.default_fonts
        ]


@dataclass
class WebsiteConfiguration:
    html_title: str = "Label Designer"
    page_title: str = "Brother QL Label Designer"
    page_headline: str = "Design your label and print it…"



from __future__ import annotations

import logging
from dataclasses import dataclass
from io import BytesIO
from typing import (
    cast,
    Literal,
    Tuple,
)  # TODO: Remove `Tuple` after dropping Python 3.8.

from brother_ql import BrotherQLRaster, create_label
from brother_ql.devicedependent import (
    ENDLESS_LABEL,
    DIE_CUT_LABEL,
    ROUND_DIE_CUT_LABEL,
    label_type_specs,
)
from brother_ql.labels import FormFactor
from brother_ql_web.configuration import Configuration
from brother_ql_web import utils
from PIL import Image, ImageDraw, ImageFont


logger = logging.getLogger(__name__)
del logging


@dataclass
class LabelParameters:
    configuration: Configuration

    font_family: str | None = None
    font_style: str | None = None
    text: str = ""
    font_size: int = 100
    label_size: str = "62"
    margin: int = 10
    threshold: int = 70
    align: str = "center"
    orientation: str = "standard"
    margin_top: int = 24
    margin_bottom: int = 45
    margin_left: int = 35
    margin_right: int = 35
    label_count: int = 1
    # TODO: Not yet taken into account. The number of dots in each direction has to be
    #       doubled. The generator/calculation methods have to be updated accordingly.
    high_quality: bool = False

    @property
    def kind(self) -> FormFactor:
        return cast(FormFactor, label_type_specs[self.label_size]["kind"])

    def _scale_margin(self, margin: int) -> int:
        return int(self.font_size * margin / 100.0)

    @property
    def margin_top_scaled(self) -> int:
        return self._scale_margin(self.margin_top)

    @property
    def margin_bottom_scaled(self) -> int:
        return self._scale_margin(self.margin_bottom)

    @property
    def margin_left_scaled(self) -> int:
        return self._scale_margin(self.margin_left)

    @property
    def margin_right_scaled(self) -> int:
        return self._scale_margin(self.margin_right)

    @property
    def fill_color(self) -> Tuple[int, int, int]:
        return (255, 0, 0) if "red" in self.label_size else (0, 0, 0)

    @property
    def font_path(self) -> str:
        try:
            if self.font_family is None or self.font_style is None:
                assert self.configuration.label.default_font is not None
                self.font_family = self.configuration.label.default_font.family
                self.font_style = self.configuration.label.default_font.style
            fonts = utils.collect_fonts(self.configuration)
            path = fonts[self.font_family][self.font_style]
        except KeyError:
            raise LookupError("Couldn't find the font & style")
        return path

    @property
    def width_height(self) -> Tuple[int, int]:
        try:
            width, height = cast(
                Tuple[int, int], label_type_specs[self.label_size]["dots_printable"]
            )
        except KeyError:
            raise LookupError("Unknown label_size")

        if height > width:
            width, height = height, width
        if self.orientation == "rotated":
            height, width = width, height
        return width, height

    @property
    def width(self) -> int:
        return self.width_height[0]

    @property
    def height(self) -> int:
        return self.width_height[1]


def _determine_image_dimensions(
    text: str, image_font: ImageFont.FreeTypeFont, parameters: LabelParameters
) -> Tuple[int, int, int, int]:
    image = Image.new("L", (20, 20), "white")
    draw = ImageDraw.Draw(image)

    left, top, right, bottom = draw.multiline_textbbox(
        xy=(0, 0), text=text, font=image_font
    )
    text_width, text_height = (right - left, bottom - top)
    width, height = parameters.width_height
    if parameters.orientation == "standard":
        if parameters.kind in (ENDLESS_LABEL,):
            height = (
                text_height
                + parameters.margin_top_scaled
                + parameters.margin_bottom_scaled
            )
    elif parameters.orientation == "rotated":
        if parameters.kind in (ENDLESS_LABEL,):
            width = (
                text_width
                + parameters.margin_left_scaled
                + parameters.margin_right_scaled
            )
    return width, height, text_width, text_height


def _determine_text_offsets(
    height: int,
    width: int,
    text_height: int,
    text_width: int,
    parameters: LabelParameters,
) -> Tuple[int, int]:
    if parameters.orientation == "standard":
        if parameters.kind in (DIE_CUT_LABEL, ROUND_DIE_CUT_LABEL):
            vertical_offset = (height - text_height) // 2
            vertical_offset += (
                parameters.margin_top_scaled - parameters.margin_bottom_scaled
            ) // 2
        else:
            vertical_offset = parameters.margin_top_scaled
        horizontal_offset = max((width - text_width) // 2, 0)
    elif parameters.orientation == "rotated":
        vertical_offset = (height - text_height) // 2
        vertical_offset += (
            parameters.margin_top_scaled - parameters.margin_bottom_scaled
        ) // 2
        if parameters.kind in (DIE_CUT_LABEL, ROUND_DIE_CUT_LABEL):
            horizontal_offset = max((width - text_width) // 2, 0)
        else:
            horizontal_offset = parameters.margin_left_scaled
    return horizontal_offset, vertical_offset


def create_label_image(parameters: LabelParameters) -> Image.Image:
    image_font = ImageFont.truetype(parameters.font_path, parameters.font_size)

    # Workaround for a bug in multiline_textsize()
    # when there are empty lines in the text:
    lines = []
    for line in parameters.text.split("\n"):
        if line == "":
            line = " "
        lines.append(line)
    text = "\n".join(lines)

    width, height, text_width, text_height = _determine_image_dimensions(
        text=text, image_font=image_font, parameters=parameters
    )
    offset = _determine_text_offsets(
        width=width,
        height=height,
        text_width=text_width,
        text_height=text_height,
        parameters=parameters,
    )

    image = Image.new("RGB", (width, height), "white")
    draw = ImageDraw.Draw(image)
    align = cast(Literal["left", "center", "right"], parameters.align)
    draw.multiline_text(
        offset, text, parameters.fill_color, font=image_font, align=align
    )
    return image


def image_to_png_bytes(image: Image.Image) -> bytes:
    image_buffer = BytesIO()
    image.save(image_buffer, format="PNG")
    image_buffer.seek(0)
    return image_buffer.read()


def generate_label(
    parameters: LabelParameters,
    configuration: Configuration,
    save_image_to: str | None = None,
) -> BrotherQLRaster:
    image = create_label_image(parameters)
    if save_image_to:
        image.save(save_image_to)

    red: bool = "red" in parameters.label_size
    rotate: int | str = 0
    if parameters.kind == ENDLESS_LABEL:
        rotate = 0 if parameters.orientation == "standard" else 90
    elif parameters.kind in (ROUND_DIE_CUT_LABEL, DIE_CUT_LABEL):
        rotate = "auto"

    if parameters.high_quality:
        logger.warning("High quality mode is not implemented for now.")

    qlr = BrotherQLRaster(configuration.printer.model)
    create_label(
        qlr,
        image,
        parameters.label_size,
        red=red,
        threshold=parameters.threshold,
        cut=True,
        rotate=rotate,
        dpi_600=False,
    )

    return qlr


def print_label(
    parameters: LabelParameters,
    qlr: BrotherQLRaster,
    configuration: Configuration,
    backend_class: type,
) -> None:
    backend = backend_class(configuration.printer.printer)
    for i in range(parameters.label_count):
        logger.info("Printing label %d of %d ...", i + 1, parameters.label_count)
        backend.write(qlr.data)
    backend.dispose()
    del backend



from __future__ import annotations

import logging
from pathlib import Path
from typing import Any, cast, Dict  # TODO: Remove `Dict` after dropping Python 3.8.

import bottle
from brother_ql_web.configuration import Configuration
from brother_ql_web.labels import (
    LabelParameters,
    create_label_image,
    image_to_png_bytes,
    generate_label,
    print_label,
)
from brother_ql_web.utils import BACKEND_TYPE


logger = logging.getLogger(__name__)
del logging

CURRENT_DIRECTORY = Path(__file__).parent


def get_config(key: str) -> object:
    return bottle.request.app.config[key]


@bottle.route("/")  # type: ignore[misc]
def index() -> None:
    bottle.redirect("/labeldesigner")


@bottle.route("/static/<filename:path>")  # type: ignore[misc]
def serve_static(filename: str) -> bottle.HTTPResponse:
    return bottle.static_file(filename, root=str(CURRENT_DIRECTORY / "static"))


@bottle.route("/labeldesigner")  # type: ignore[misc]
@bottle.jinja2_view("labeldesigner.jinja2")  # type: ignore[misc]
def labeldesigner() -> dict[str, Any]:
    fonts = cast(Dict[str, Dict[str, str]], get_config("brother_ql_web.fonts"))
    font_family_names = sorted(list(fonts.keys()))
    configuration = cast(Configuration, get_config("brother_ql_web.configuration"))
    return {
        "font_family_names": font_family_names,
        "fonts": fonts,
        "label_sizes": get_config("brother_ql_web.label_sizes"),
        "website": configuration.website,
        "label": configuration.label,
        "default_orientation": configuration.label.default_orientation,
    }


def get_label_parameters(request: bottle.BaseRequest) -> LabelParameters:
    # As we have strings, *bottle* would try to generate Latin-1 bytes from it
    # before decoding it back to UTF-8. This seems to break some umlauts, thus
    # resulting in UnicodeEncodeErrors being raised when going back to UTF-8.
    # For now, we just state that we always receive clean UTF-8 data and thus
    # the recode operations just can be omitted. All external API users are
    # responsible for passing clean data.
    # References:
    #   * https://github.com/bottlepy/bottle/blob/99341ff3791b2e7e705d7373e71937e9018eb081/bottle.py#L2197-L2203  # noqa: E501
    #   * https://github.com/FriedrichFroebel/brother_ql_web/issues/9
    parameters = request.params
    parameters.recode_unicode = False
    d = parameters.decode()  # UTF-8 decoded form data

    font_family = d.get("font_family").rpartition("(")[0].strip()
    font_style = d.get("font_family").rpartition("(")[2].rstrip(")")
    context = {
        "text": d.get("text", ""),
        "font_size": int(d.get("font_size", 100)),
        "font_family": font_family,
        "font_style": font_style,
        "label_size": d.get("label_size", "62"),
        "margin": int(d.get("margin", 10)),
        "threshold": int(d.get("threshold", 70)),
        "align": d.get("align", "center"),
        "orientation": d.get("orientation", "standard"),
        "margin_top": int(d.get("margin_top", 24)),
        "margin_bottom": int(d.get("margin_bottom", 45)),
        "margin_left": int(d.get("margin_left", 35)),
        "margin_right": int(d.get("margin_right", 35)),
        "label_count": int(d.get("label_count", 1)),
        "high_quality": bool(d.get("high_quality", False)),  # TODO: Enable by default.
        "configuration": request.app.config["brother_ql_web.configuration"],
    }

    return LabelParameters(**context)


@bottle.get("/api/preview/text")  # type: ignore[misc]
@bottle.post("/api/preview/text")  # type: ignore[misc]
def get_preview_image() -> bytes:
    parameters = get_label_parameters(bottle.request)
    image = create_label_image(parameters=parameters)
    return_format = bottle.request.query.get("return_format", "png")
    if return_format == "base64":
        import base64

        bottle.response.set_header("Content-type", "text/plain")
        return base64.b64encode(image_to_png_bytes(image))
    else:
        bottle.response.set_header("Content-type", "image/png")
        return image_to_png_bytes(image)


@bottle.post("/api/print/text")  # type: ignore[misc]
@bottle.get("/api/print/text")  # type: ignore[misc]
def print_text() -> dict[str, bool | str]:
    """
    API to print a label

    returns: JSON
    """
    return_dict: dict[str, bool | str] = {"success": False}

    try:
        parameters = get_label_parameters(bottle.request)
    except (AttributeError, LookupError) as e:
        return_dict["error"] = str(e)
        return return_dict

    if parameters.text is None:
        return_dict["error"] = "Please provide the text for the label"
        return return_dict

    qlr = generate_label(
        parameters=parameters,
        configuration=cast(Configuration, get_config("brother_ql_web.configuration")),
        save_image_to="sample-out.png" if bottle.DEBUG else None,
    )

    if not bottle.DEBUG:
        try:
            print_label(
                parameters=parameters,
                qlr=qlr,
                configuration=cast(
                    Configuration, get_config("brother_ql_web.configuration")
                ),
                backend_class=cast(
                    BACKEND_TYPE,
                    get_config("brother_ql_web.backend_class"),
                ),
            )
        except Exception as e:
            return_dict["message"] = str(e)
            logger.warning("Exception happened: %s", e)
            return return_dict

    return_dict["success"] = True
    if bottle.DEBUG:
        return_dict["data"] = str(qlr.data)
    return return_dict


def main(
    configuration: Configuration,
    fonts: dict[str, dict[str, str]],
    label_sizes: list[tuple[str, str]],
    backend_class: BACKEND_TYPE,
) -> None:
    app = bottle.default_app()
    app.config["brother_ql_web.configuration"] = configuration
    app.config["brother_ql_web.fonts"] = fonts
    app.config["brother_ql_web.label_sizes"] = label_sizes
    app.config["brother_ql_web.backend_class"] = backend_class
    bottle.TEMPLATE_PATH.append(CURRENT_DIRECTORY / "views")
    debug = configuration.server.is_in_debug_mode
    app.run(host=configuration.server.host, port=configuration.server.port, debug=debug)
